from widgets import *
from log import gettimezone
import re, time

ROWSIZES = [4, 10, 40, 4]

class Row(Frame):
    def __init__(self, parent, editcommand=None, bg='white', grid='#d9d9d9'):
        Frame.__init__(self, parent, background=grid)
        Entry = curry(AquaEntry, background=bg, command=editcommand)
        self.start = Entry(self, width=5, justify=RIGHT)
        self.start.pack(side=LEFT)
        self.action = Entry(self)
        self.action.pack(side=LEFT, padx=1, fill=X, expand=1)
        self.duration = Entry(self, width=4, justify=RIGHT)
        self.duration.pack(side=RIGHT)

class Input(Frame):
    def __init__(self, parent, editcommand=None, togglecommand=None):
        Frame.__init__(self, parent, borderwidth=0, relief=SUNKEN)
        self.action = AquaEntry(
            self, font='{Lucida Grande} 12', command=editcommand)
        self.action.pack(side=LEFT, padx=1, fill=X, expand=1)
        self.toggle = Toggle(self, command=togglecommand)
        self.toggle.pack(side=RIGHT)

UPDATE_PERIOD = 2000 # ms

def parsetime(text):
    text = text.replace(';', ':')
    if ':' in text:
        hours, minutes = text.split(':', 1)
    else:
        hours, minutes = 0, text
    return int(hours or 0), int(minutes or 0)

def neartime(ref, h, m):
    y, l, d, _, _, _, _, _, dst = time.localtime(ref)
    near = time.mktime((y, l, d, h, m, 0, 0, 0, dst))
    while near > ref + 12*3600:
        near -= 24*3600
    while near < ref - 12*3600:
        near += 24*3600
    return near

class Main(Toplevel):
    def __init__(self, parent, log):
        Toplevel.__init__(self, parent)
        self.transient(parent)
        self.log = log
        self.log.listener = self.updatelogview
        self.tk.createtimerhandler(UPDATE_PERIOD, self.tick)
        self.lasttick = time.time()
        self.deactivated = None
        self.initmenus()
        self.rows = []
        self.rowframe = f = Frame(self)
        bg = '#ffffff'
        for i in range(-max(ROWSIZES), 0):
            bg = (bg == '#ffffff') and '#edf3fe' or '#ffffff'
            row = Row(f, lambda w, o, n, i=i: self.editrow(i, w, o, n), bg)
            self.rows.append(row)
        for prev, row, next in zip(
            [None] + self.rows[:-1], self.rows, self.rows[1:] + [None]):
            if prev:
                row.start.bind('<Key-Up>', prev.start.focus)
                row.action.bind('<Key-Up>', prev.action.focus)
                row.duration.bind('<Key-Up>', prev.duration.focus)
            if next:
                row.start.bind('<Key-Down>', next.start.focus)
                row.action.bind('<Key-Down>', next.action.focus)
                row.duration.bind('<Key-Down>', next.duration.focus)
        self.input = Input(self, self.edit, self.toggle)
        self.expanded = 0
        self.nrows = ROWSIZES[0]
        self.show(self.nrows)
        self.bind('<Command-Key>q', self.close)
        self.bind('<Command-Key>e', self.toggle)
        self.bind('<Command-Key>l', self.focus)
        self.bind('<Command-Key>E', self.togglerows)
        self.bind('<Activate>', self.activate)
        self.bind('<Deactivate>', self.deactivate)
        if log and not log[-1][-1].strip().endswith('.'):
            self.input.action.set(log[-1][-1])
        self.focus()
        self.timezone = gettimezone()
        self.updatelogview()
        self.update()

    def updatelogview(self):
        count = len(self.rows)
        entries = self.log[-count:]
        entries = ([None] * count + entries)[-count:]
        now = time.time()
        for i in range(count):
            if entries[i]:
                start, action = entries[i]
                end = (i + 1 < count) and entries[i + 1][0] or now
                duration = int((end - start)/60)
                _, _, _, h, m, _, _, _, _ = time.localtime(start)
                start = '%02d:%02d' % (h, m)
                duration = duration and '%d:%02d' % (duration/60, duration%60)
            else:
                start = duration = action = end = ''
            self.rows[i].start.set(start)
            self.rows[i].duration.set(duration or '')
            self.rows[i].action.set(action)

    def tick(self):
        now = time.time()
        if now > self.lasttick + 30:
            if self.log and self.log[-1][1].endswith('.'):
                self.log.append((self.lasttick, ''))
                self.log.save()
                self.updatelogview()
        elif self.log:
            start, action = self.log[-1]
            duration = int((now - start)/60)
            duration = duration and '%d:%02d' % (duration/60, duration%60)
            self.rows[-1].duration.set(duration or '')
        if self.timezone != gettimezone():
            self.timezone = gettimezone()
            self.updatelogview()
        if self.deactivated and now > self.deactivated + 10:
            if self.expanded:
                self.toggle()
            sw, sh = self.winfo_screenwidth(), self.winfo_screenheight()
            self.geometry('+%d+%d' % (sw, sh))
        self.tk.createtimerhandler(UPDATE_PERIOD, self.tick)
        self.lasttick = now

    def close(self, *args):
        self.quit()

    def activate(self, *args):
        self.deiconify()
        self.update()
        self.deactivated = None

    def deactivate(self, *args):
        self.deactivated = time.time()

    def focus(self, *args):
        self.input.action.focus()
        self.input.action.selectall()
        self.update()
        self.input.action.focus()
        self.input.action.selectall()

    def edit(self, widget, oldvalue, newvalue):
        if re.match(r'\d\d?[:;]\d\d? ', newvalue):
            start, newvalue = newvalue.split(None, 1)
            start = neartime(self.log[-1][0], *parsetime(start))
            self.input.action.set(newvalue)
        elif re.match(r'\+\d+ ', newvalue):
            delta, newvalue = newvalue.split(None, 1)
            start = self.log[-1][0] + int(delta)*60
            self.input.action.set(newvalue)
        elif re.match(r'-\d+ ', newvalue):
            delta, newvalue = newvalue.split(None, 1)
            start = time.time() + int(delta)*60
            self.input.action.set(newvalue)
        else:
            start = time.time()
        entry = (start, newvalue)
        if self.log and start < self.log[-1][0] + 60:
            self.log[-1] = entry
        else:
            self.log.append(entry)
        self.log.save()

    def editrow(self, index, widget, oldvalue, newvalue):
        row = self.rows[index]
        now = time.time()
        start, action = self.log[index]
        end = index < -1 and self.log[index + 1][0] or time.time()
        duration = int((end - start)/60)
        if widget is row.start:
            if re.match('r[-+]\d+', newvalue):
                row.start.set(oldvalue + int(newvalue)*60)
            else:
                try:
                    h, m = parsetime(newvalue)
                except ValueError:
                    row.start.set(oldvalue)
                else:
                    newstart = neartime(start, h, m)
                    newstart = min(newstart, now)
                    if -len(self.log) < index:
                        newstart = max(newstart, self.log[index-1][0])
                    self.log[index] = (newstart, self.log[index][1])
                    for i in range(index + 1, 0):
                        if newstart > self.log[i][0]:
                            self.log[i] = (newstart, self.log[i][1])
            row.start.selectall()
        if widget is row.action:
            self.log[index] = (start, newvalue)
            if index == -1:
                self.input.action.set(newvalue)
        if widget is row.duration:
            if re.match('r[-+]\d+', newvalue):
                row.duration.set(oldvalue + int(newvalue)*60)
            else:
                try:
                    h, m = parsetime(newvalue)
                except ValueError:
                    row.duration.set(oldvalue)
                else:
                    newstart = min(start + h*3600 + m*60, now)
                    if index < -1:
                        self.log[index + 1] = (newstart, self.log[index + 1][1])
                        for i in range(index + 2, 0):
                            if newstart > self.log[i][0]:
                                self.log[i] = (newstart, self.log[i][1])
            row.duration.selectall()
        self.updatelogview()
        self.log.save()

    def toggle(self, *args):
        self.expanded = not self.expanded
        self.viewmenu.entryconfigure(
            0, label=self.expanded and 'Collapse' or 'Expand')
        self.input.toggle.pushed = self.expanded
        self.input.toggle.update()
        self.focus()

    def show(self, nrows):
        for row in self.rows:
            row.forget()
        for row in self.rows[-nrows:]:
            row.pack(side=TOP, fill=X, expand=1)

    def togglerows(self, *args):
        self.nrows = ROWSIZES[ROWSIZES.index(self.nrows) + 1]
        self.show(self.nrows)
        self.position()

    def update(self):
        self.input.pack(side=BOTTOM, fill=X, expand=1)
        if self.expanded:
            self.rowframe.pack(fill=X, expand=1)
        else:
            self.rowframe.forget()
        self.position()

    def position(self):
        sw, sh = self.winfo_screenwidth(), self.winfo_screenheight()
        # self.geometry('+%d+%d' % (sw, sh))
        # Toplevel.update(self)
        w = 300
        h = self.input.winfo_reqheight()
        if self.expanded:
            for row in self.rows[-self.nrows:]:
                h += row.winfo_reqheight()
        self.geometry('%dx%d+%d+%d' % (w, h, sw-w-1, 21))

    def initmenus(self):
        menubar = Menu(self)
        self['menu'] = menubar
        self.viewmenu = Menu(menubar)
        menubar.add_cascade(menu=self.viewmenu, label='View')
        self.viewmenu.add_command(
            label='Expand', accelerator='Command-E', command=self.toggle)
